/*
 * Translate Prolog format CSP to ASP(Order Encoding)
 * @version 2.1.4
 * @date 2013-12-18
 */
:- dynamic clause_count/1.

trans(InFile) :-
	open(InFile, read, In),
	repeat,
  	  read(In, X),
	  translate(X),
	X == end_of_file,
	close(In).

translate(end_of_file) :- !.
translate(X) :-
	tr(X, Ss, []),
	!,
	output(Ss).
translate(X) :-
	output([error(X)]).

output([]).
output([S|Ss]) :-
	writeClause(S), write('.'), nl,
	output(Ss).

writeClause(or(X)) :- !,
        write('1 {'), writeClause(X), write('}').
writeClause(not(A)) :- !,
        write('not '), writeClause(A).
writeClause(nodom(A,range(B))) :- !,
        writeClause(nodom(A,B)).
writeClause(nodom(A,range(B,C))) :- !,
        write('nodom('), writeClause(A), write(','),
        writeClause(B), write('..'), writeClause(C), write(')').
writeClause([C]) :- !,
        writeClause(C).
writeClause([C|Cs]) :- !,
        writeClause(C), write(', '), writeClause(Cs).
writeClause(C) :- !,
        writeq(C).

comment(C) :-
        write('% '), write(C), nl.

tr([int,V,Domain]) -->
	tr_int(V, Domain).
tr([bool,V]) -->
	tr_bool(V).
tr([not,V]) -->
	tr_not(V).
tr([relation,R,N,Body]) -->
	tr_relation(R, N, Body).
tr([objective,Target,V]) -->
	[objective(Target,V)].
tr(Clause) -->
	tr_clause(Clause).

/*
 * Integer Variable
 */
tr_int(V, [D|Ds]) -->
%        comment('int V Domain'),
        {tr_domain(D, L, U)},
        tr_ints(V, L, U, Ds).
tr_ints(V, LB, UB, []) -->
        [var(V), lb(V, LB), ub(V, UB)].
tr_ints(V, LB, A, [X|Xs]) -->
        {tr_domain(X, L, U)},
        tr_ints(V, LB, U, Xs),
        {A1 is A + 1},
        {L1 is L - 1},
        [nodom(V, range(A1,L1))].

        
/*
 * Boolean Variable
 */
tr_bool(V) -->
	[bool(V)].

/*
 * not
 */
tr_not(V) -->
	[:-  V].

/*
 * Relation
 */
tr_relation(R, N, [Type|Tuples]) -->
	[relation(R, N, Type, Tuples)].

/*
 * Clause
 */
% tr_clause([or|Lits]) --> !,
% 	{clause_number(N)},
% 	[constraint(N)],
% 	tr_literals(N, Lits).
% tr_clause(Lit) -->
%  	tr_clause([or,Lit]).
tr_clause(Lit) -->
	tr_literals(Lit).

clause_number(N) :-
	(retract(clause_count(N0)) ; N0 = 0),
	!,
	N is N0 + 1,
	assert(clause_count(N)).

tr_literals(_, []) --> !.
tr_literals(N, [Lit|Lits]) -->
	{tr_literal(Lit, Z)},
	[constraint_elem(N, Z)],
%	[Z],
	tr_literals(N, Lits).
tr_literals([]) --> !.
tr_literals(Lit) -->
	{tr_literal(Lit, Z)},
	[Z].
%	tr_literals(Lits).

tr_literal(Lit, Lit) :- atomic(Lit), !.
tr_literal([not, Lit], Z) :- !,
	tr_literal(Lit, X),
        Z = not(X).
tr_literal([or, X], Z) :- !,
        tr_literal(X, Z).
tr_literal([or|X], Z) :- !,
        comment('or '),
        tr_or(X, Xs),
        Z = or(Xs).

%tr_literal([or, X1, X2], Z) :- !,
%        comment('or X1 X2'),
%        tr_literal(X1, Z1),
%        tr_literal(X2, Z2),
%        Z = or(Z1, Z2).
%tr_literal([or, X1, X2, X3], Z) :- !,
%        comment('or X1 X2 X3'),
%        tr_literal(X1, Z1),
%        tr_literal(X2, Z2),
%        tr_literal(X3, Z3),
%        Z = or(Z1, Z2, Z3).
tr_literal([wsum, [[A1,X1]], le, B], Z) :- !,
        comment('A1 X1 <= B'),
        Z = le0(A1,X1,B).
tr_literal([wsum, [[A1,X1],[A2,X2]], le, B], Z) :- !,
        comment('A1 X1 + A2 X2 <= B'),
        Z = le(A1,X1,A2,X2,B).
tr_literal([wsum, [[A1,X1],[A2,X2],[A3,X3]], le, B], Z) :- !,
        comment('A1 X1 + A2 X2 + A3 X3 <= B'),
        Z = le(A1,X1,A2,X2,A3,X3,B).
tr_literal([wsum, [[A1,X1]], ge, B], Z) :- !,
        comment('A1 X1 >= B'),
        Z = ge0(A1,X1,B).
tr_literal([wsum, [[A1,X1],[A2,X2]], ge, B], Z) :- !,
        comment('A1 X1 + A2 X2 >= B'),
        Z = ge(A1,X1,A2,X2,B).
tr_literal([wsum, [[A1,X1],[A2,X2],[A3,X3]], ge, B], Z) :- !,
        comment('A1 X1 + A2 X2 + A3 X3 >= B'),
        Z = ge(A1,X1,A2,X2,A3,X3,B).

tr_literal([wsum, AXs, Term, B], Z) :- !,
        append(AXs, B ,AX1),
        flatten(AX1, AX2),
        Z = le(AX2, Term).
%        tr_list(AX2, Term, Z).
%        tr_list(AX1, Term, AX2),
%        Z = le(AX2, Term, B).
%	tr_wsum(AXs, Term, B, Z).

tr_literal([alldifferent, Xs], Z) :- !,
	Z = alldifferent(Xs).
tr_literal([weightedsum, AXs, Comp, B], Z) :- !,
	tr_list(AXs, mul, AXs1),
	Z = wsum(AXs1, Comp, B).
tr_literal([cumulative, Tasks, Limit], Z) :- !,
	tr_list(Tasks, task, Tasks1),
	Z = cumulative(Tasks1, Limit).
tr_literal([element, Index, Terms, Value], Z) :- !,
	Z = element(Index, Terms, Value).
tr_literal([djsjunctive, Tasks], Z) :- !,
	tr_list(Tasks, task, Tasks1),
	Z = disjunctive(Tasks1).
tr_literal([lex_less, Xs, Ys], Z) :- !,
	Z = lex_less(Xs, Ys).
tr_literal([lex_lesseq, Xs, Ys], Z) :- !,
	Z = lex_lesseq(Xs, Ys).
tr_literal([nvalue, Count, Xs], Z) :- !,
	Z = nvalue(Count, Xs).
tr_literal([global_cardinality, Xs, Cards], Z) :- !,
	tr_list(Cards, card, Cards1),
	Z = global_cardinality(Xs, Cards1).
tr_literal([global_cardinality_with_costs, Xs, Cards, Costs, Cost], Z) :- !,
	tr_list(Cards, card, Cards1),
	tr_list(Costs, cost, Costs1),
	Z = global_cardinality_with_costs(Xs, Cards1, Costs1, Cost).
tr_literal([count, X, Xs, Comp, Count], Z) :- !,
	Z = count(X, Xs, Comp, Count).
%tr_literal([R|Xs], Z) :-
%        comment(R),
%	Z = rel(R, Xs).
tr_literal([X|Xs], Z) :-
        Z = X, Xs.

tr_wsum_le([[A,X]], B, Z) :- A > 0, !,
	C is integer(floor(B/A)),
        comment('AX <= B (A > 0)'),
	Z = le0(X, C).
tr_wsum_le([[A,X]], B, Z) :- A < 0, !,
	C is integer(ceiling(B/A)),
        comment('AX <= B (A < 0)'),
	Z = ge0(X, C).
tr_wsum_le(AXs, B, Z) :-
        comment('A1X + A2X + ... <= B'),
	tr_list(AXs, AXs1),
        flatten(AXs1, AXs2),
        concat_atom(AXs2, ',',AXs3),
        Z = le(AXs3, B).

tr_wsum([X|Xs], F, [Z|Zs]) :-
        Z =.. [F|X],
        tr_wsum(Xs, F, Zs).
tr_wsum([[A,X]], Term, B, Z) :-
	C is integer(floor(B/A)),
        Zz = [X, C],
	Z =.. [Term|Zz].

tr_list([], _, []).
tr_list([X|Xs], F, [Z|Zs]) :-
	Z =.. [F|X],
	tr_list(Xs, F, Zs).

tr_list([], []).
tr_list([X|Xs], [Z|Zs]) :-
	Z = X,
	tr_list(Xs, Zs).

tr_domain([], 0, 0).
tr_domain([LB,UB], LB, UB).

tr_or([], []).
tr_or([X|Xs], [Z|Zs]) :- 
        tr_literal(X, Z),
        tr_or(Xs, Zs).

%tr_join([], []).
%tr_join([X|Xs], [X,",",Zs]) :-
%        tr_join(Xs, Zs).
